@@ -2,6 +2,7 @@ package convert
 
 import (
 	"fmt"
+	"math/big"
 	"testing"
 
 	"github.com/zclconf/go-cty/cty"
@@ -143,6 +144,14 @@ func TestConvert(t *testing.T) {
 			Type:      cty.List(cty.DynamicPseudoType),
 			WantError: true, // there is no type that both tuple elements can unify to for conversion to list
 		},
+		{
+			Value: cty.SetVal([]cty.Value{
+				cty.StringVal("5"),
+				cty.UnknownVal(cty.String),
+			}),
+			Type: cty.Set(cty.Number),
+			Want: cty.SetVal([]cty.Value{cty.NumberIntVal(5), cty.UnknownVal(cty.Number)}),
+		},
 		{
 			Value: cty.SetVal([]cty.Value{
 				cty.StringVal("5"),
@@ -182,6 +191,27 @@ func TestConvert(t *testing.T) {
 				cty.StringVal("10"),
 			}),
 		},
+		{
+			Value: cty.SetVal([]cty.Value{
+				cty.StringVal("5"),
+				cty.UnknownVal(cty.String),
+			}),
+			Type: cty.List(cty.String),
+			Want: cty.UnknownVal(cty.List(cty.String)),
+		},
+		{
+			Value: cty.SetVal([]cty.Value{
+				cty.UnknownVal(cty.String),
+			}),
+			Type: cty.List(cty.String),
+			// We get a known list value this time because even though we
+			// don't know the single value that's in the list, we _do_ know
+			// that there are no other values in the set for it to coalesce
+			// with.
+			Want: cty.ListVal([]cty.Value{
+				cty.UnknownVal(cty.String),
+			}),
+		},
 		{
 			Value: cty.ListVal([]cty.Value{
 				cty.NumberIntVal(5),
@@ -381,6 +411,32 @@ func TestConvert(t *testing.T) {
 				"name": cty.StringVal("John"),
 			}),
 		},
+		{
+			Value: cty.MapVal(map[string]cty.Value{
+				"name": cty.StringVal("John"),
+			}),
+			Type: cty.Object(map[string]cty.Type{
+				"name":     cty.String,
+				"greeting": cty.String,
+			}),
+			WantError: true, // map has no element for required attribute "greeting"
+		},
+		{
+			Value: cty.MapVal(map[string]cty.Value{
+				"name": cty.StringVal("John"),
+			}),
+			Type: cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"name":     cty.String,
+					"greeting": cty.String,
+				},
+				[]string{"greeting"},
+			),
+			Want: cty.ObjectVal(map[string]cty.Value{
+				"greeting": cty.NullVal(cty.String),
+				"name":     cty.StringVal("John"),
+			}),
+		},
 		{
 			Value: cty.MapVal(map[string]cty.Value{
 				"a": cty.NumberIntVal(2),
@@ -476,6 +532,89 @@ func TestConvert(t *testing.T) {
 			}),
 			WantError: true, // given value must have superset object type
 		},
+		{
+			Value: cty.ObjectVal(map[string]cty.Value{
+				"bar": cty.StringVal("bar value"),
+			}),
+			Type: cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				},
+				[]string{"foo"},
+			),
+			Want: cty.ObjectVal(map[string]cty.Value{
+				"foo": cty.NullVal(cty.String),
+				"bar": cty.StringVal("bar value"),
+			}),
+		},
+		{
+			Value: cty.ObjectVal(map[string]cty.Value{
+				"foo": cty.StringVal("foo value"),
+				"bar": cty.StringVal("bar value"),
+			}),
+			Type: cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				},
+				[]string{"foo"},
+			),
+			Want: cty.ObjectVal(map[string]cty.Value{
+				"foo": cty.StringVal("foo value"),
+				"bar": cty.StringVal("bar value"),
+			}),
+		},
+		{
+			Value: cty.EmptyObjectVal,
+			Type: cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				},
+				[]string{"foo"},
+			),
+			WantError: true, // Attribute "bar" is required
+		},
+		{
+			Value: cty.NullVal(cty.DynamicPseudoType),
+			Type: cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				},
+				[]string{"foo"},
+			),
+			Want: cty.NullVal(cty.Object(map[string]cty.Type{
+				"foo": cty.String,
+				"bar": cty.String,
+			})),
+		},
+		{
+			Value: cty.ListVal([]cty.Value{
+				cty.NullVal(cty.DynamicPseudoType),
+				cty.ObjectVal(map[string]cty.Value{
+					"bar": cty.StringVal("bar value"),
+				}),
+			}),
+			Type: cty.List(cty.ObjectWithOptionalAttrs(
+				map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				},
+				[]string{"foo"},
+			)),
+			Want: cty.ListVal([]cty.Value{
+				cty.NullVal(cty.Object(map[string]cty.Type{
+					"foo": cty.String,
+					"bar": cty.String,
+				})),
+				cty.ObjectVal(map[string]cty.Value{
+					"foo": cty.NullVal(cty.String),
+					"bar": cty.StringVal("bar value"),
+				}),
+			}),
+		},
 		{
 			Value: cty.ObjectVal(map[string]cty.Value{
 				"foo": cty.True,
@@ -649,6 +788,92 @@ func TestConvert(t *testing.T) {
 				"b": cty.MapValEmpty(cty.String),
 			}),
 		},
+		// reduction of https://github.com/hashicorp/terraform/issues/27269
+		{
+			Value: cty.TupleVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"a": cty.NullVal(cty.DynamicPseudoType),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"a": cty.ObjectVal(map[string]cty.Value{
+						"b": cty.ListVal([]cty.Value{
+							cty.ObjectVal(map[string]cty.Value{
+								"c": cty.StringVal("d"),
+							}),
+						}),
+					}),
+				}),
+			}),
+			Type: cty.List(cty.Object(map[string]cty.Type{
+				"a": cty.Object(map[string]cty.Type{
+					"b": cty.List(cty.ObjectWithOptionalAttrs(map[string]cty.Type{
+						"c": cty.String,
+						"d": cty.String,
+					}, []string{"d"})),
+				}),
+			})),
+			Want: cty.ListVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"a": cty.NullVal(cty.Object(map[string]cty.Type{
+						"b": cty.List(cty.Object(map[string]cty.Type{
+							"c": cty.String,
+							"d": cty.String,
+						})),
+					})),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"a": cty.ObjectVal(map[string]cty.Value{
+						"b": cty.ListVal([]cty.Value{
+							cty.ObjectVal(map[string]cty.Value{
+								"c": cty.StringVal("d"),
+								"d": cty.NullVal(cty.String),
+							}),
+						}),
+					}),
+				}),
+			}),
+		},
+		// When converting null values into nested types which include objects
+		// with optional attributes, we expect the resulting value to be of a
+		// recursively concretized type.
+		{
+			Value: cty.NullVal(cty.DynamicPseudoType),
+			Type: cty.Object(
+				map[string]cty.Type{
+					"foo": cty.ObjectWithOptionalAttrs(
+						map[string]cty.Type{
+							"bar": cty.String,
+						},
+						[]string{"bar"},
+					),
+				},
+			),
+			Want: cty.NullVal(cty.Object(map[string]cty.Type{
+				"foo": cty.Object(map[string]cty.Type{
+					"bar": cty.String,
+				}),
+			})),
+		},
+		// The same nested optional attributes flattening should happen for
+		// unknown values, too.
+		{
+			Value: cty.UnknownVal(cty.DynamicPseudoType),
+			Type: cty.Object(
+				map[string]cty.Type{
+					"foo": cty.ObjectWithOptionalAttrs(
+						map[string]cty.Type{
+							"bar": cty.String,
+						},
+						[]string{"bar"},
+					),
+				},
+			),
+			Want: cty.UnknownVal(cty.Object(map[string]cty.Type{
+				"foo": cty.Object(map[string]cty.Type{
+					"bar": cty.String,
+				}),
+			})),
+		},
 		// https://github.com/hashicorp/terraform/issues/21588:
 		{
 			Value: cty.TupleVal([]cty.Value{
@@ -766,6 +991,114 @@ func TestConvert(t *testing.T) {
 			}),
 			WantError: false,
 		},
+		{
+			Value: cty.ListVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.ListVal([]cty.Value{
+						cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.ListValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			Type: cty.List(cty.Object(map[string]cty.Type{
+				"xs": cty.List(cty.ObjectWithOptionalAttrs(map[string]cty.Type{
+					"x": cty.Number,
+				}, []string{"x"})),
+			})),
+			Want: cty.ListVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.ListVal([]cty.Value{
+						cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.ListValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			WantError: false,
+		},
+		{
+			Value: cty.SetVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.SetVal([]cty.Value{
+						cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.SetValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			Type: cty.Set(cty.Object(map[string]cty.Type{
+				"xs": cty.Set(cty.ObjectWithOptionalAttrs(map[string]cty.Type{
+					"x": cty.Number,
+				}, []string{"x"})),
+			})),
+			Want: cty.SetVal([]cty.Value{
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.SetVal([]cty.Value{
+						cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.SetValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			WantError: false,
+		},
+		{
+			Value: cty.MapVal(map[string]cty.Value{
+				"foo": cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.MapVal(map[string]cty.Value{
+						"nested_foo": cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				"bar": cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.MapValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			Type: cty.Map(cty.Object(map[string]cty.Type{
+				"xs": cty.Map(cty.ObjectWithOptionalAttrs(map[string]cty.Type{
+					"x": cty.Number,
+				}, []string{"x"})),
+			})),
+			Want: cty.MapVal(map[string]cty.Value{
+				"foo": cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.MapVal(map[string]cty.Value{
+						"nested_foo": cty.ObjectVal(map[string]cty.Value{
+							"x": cty.NumberVal(big.NewFloat(1234)),
+						}),
+					}),
+				}),
+				"bar": cty.ObjectVal(map[string]cty.Value{
+					"xs": cty.MapValEmpty(cty.Object(map[string]cty.Type{
+						"x": cty.Number,
+					})),
+				})},
+			),
+			WantError: false,
+		},
 	}
 
 	for _, test := range tests {
